<?php

/**
 * PHPOF2/DB.php - abstract database functions and make them easier to use
 *
 * Copyright 2001-2008 Tim Jackson (tim@timj.co.uk)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of version 3 of the GNU Lesser General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @package PHPOF2
 *
 */

require_once 'MDB2.php';
require_once 'PHPOF2/Driver/DBResult/mysqli.php'; // HACK:For result_class
require_once 'PHPOF2/Driver/DBRow/mysqli.php';
require_once 'PHPOF2/DBTable.php';

/**
 * As well as the standard MDB2 options, PHPOF2 provides the following:
 *  - audit_driver - string|PHPOF2_Audit_Interface  Either an instantiated
 *                   auditor object which implements PHPOF2_Audit_Interface OR
 *                   the name of a class to load which implements
 *                    PHPOF2_Audit_Interface
 *  - audit_driver_options - array  Only applicable if the "audit_driver" option
 *                           above is a string, this is an array of options to
 *                           pass to the audit driver class which will be 
 *                           instantiated
 */
class PHPOF2_DB extends MDB2 {

	const LOB_PLACEHOLDER =   '!lob-data-present!';
	const LOB_DELETEPENDING = '!lob-delete-pending!';
	
	const KEY_PRI = 'pri';

	function &factory($dsn, $options = false)
	{
		// Options that are specific to PHPOF2_DB and are not to be passed to MDB2
		// (MDB2 throws errors if it gets options it doesn't know about)
		$private_option_list = array('audit_driver', 'audit_driver_options');
		$private_options = array();
		if (is_array($options)) {
			foreach ($options as $key => $val) {
				if (in_array($key, $private_option_list)) {
					$private_options[$key] = $val;
					unset($options[$key]);
				}
			}
		}
		
		preg_match('|^(.+)://|', $dsn, $matches);
		require_once 'PHPOF2/Driver/DBResult/' . $matches[1] . '.php';
		$options['result_class'] = 'PHPOF2_Driver_DBResult_%s';
		$options['buffered_result_class'] = 'PHPOF2_Driver_DBResult_Buffered_%s';
		$options['idxname_format'] = '%s';   // Stop the default of checking for a postfix '_idx'
		
		// Load driver-specific optimisations, if available
		@include_once 'PHPOF2/Optimisation/' . $matches[1] . '.php';
		
		// Disable MDB2 portability mode unless the user explicitly asks for it.
		// It causes lots of weird problems like broken binaries, null/empty
		// confusion etc.
		if (!isset($options['portability'])) {
			$options['portability'] = MDB2_PORTABILITY_NONE;
		}
		
		// The factory connection method (parent::factory($dsn, $options)) will 
		// use JIT connects. Using connect will force the connection to connect 
		// now, which is better for error handling.
		$mdb2 = parent::connect( $dsn, $options );
		if (PEAR::isError($mdb2)) {
    		throw new Exception($mdb2->getMessage());
		}
		
		// Enable audit driver if required
		if (isset($private_options['audit_driver'])) {		
			// If the driver is a string try and load the class
			if (is_string($private_options['audit_driver'])) {
				$driver = $private_options['audit_driver'];
				if (isset($private_options['audit_driver_options'])) {
					if (is_array($private_options['audit_driver_options'])) {
						$mdb2->phpof2_auditor = new $driver($mdb2, $private_options['audit_driver_options']);
					} else {
						throw new Exception('Audit driver options passed to PHPOF2_DB are not an array');
					}
				} else {
					$mdb2->phpof2_auditor = new $driver($mdb2);
				}
			} else {
				$mdb2->phpof2_auditor = $private_options['audit_driver'];
			}
			
			if (!$mdb2->phpof2_auditor instanceof PHPOF2_Audit_Interface) {
				throw new Exception("Audit driver passed to PHPOF2_DB is not a PHPOF2_Audit_Interface object");
			}
		}
		return $mdb2;
	}
}
